import os
import pickle
import pandas as pd
from . import feature_selection

PATRIC_FILE_EXTENSION_TO_PGFAM_COL = {'.txt' : 'pgfam', '.tab' : 'pgfam_id'}
GENOME_ID = 'Genome ID'
LABEL = 'Label'
HP = 'HP'
NHP = 'NHP'


def read_merged_file(file_path):
    """
    Reads genomes merged file into pd.Series object

    Parameters
    ----------
    file_path - path to a merged input *.fasta file

    Returns
    ----------
    pd.Series object that represents the all input genomes of the merged file
    """

    genomes_order = []
    genome_to_pgfams = {}

    with open(file_path) as f:
        genome_id = ''
        for line in f:
            if line.startswith('>'):
                genome_id = line.strip()[1:]
                genomes_order.append(genome_id)
            else:
                pgfam_id = line.strip()
                genome_to_pgfams.setdefault(genome_id, []).append(pgfam_id)

    genomes_pgfams = [' '.join(genome_to_pgfams[genome]) for genome in genomes_order]

    return pd.Series(genomes_pgfams, index=genomes_order, dtype="string")


def read_genome_file(file_entry, pgfam_col):
    """
    Reads a single genome file and returns its contained pgfams

    Parameters
    ----------
    file_entry - entry to an input genome file

    Returns
    ----------
    pd.Series object that represents all the input genomes in the directory
    """

    pgfams = pd.read_csv(file_entry, usecols=[pgfam_col], sep='\t').dropna()
    pgfams = ' '.join(list(pgfams[pgfam_col]))

    return pgfams


def read_files_in_dir(dir_path):
    """
    Reads all genomes *.txt/*.tab files in a directory into pd.Series object

    Parameters
    ----------
    dir_path - a path to an input directory with genome *.txt/*.tab files

    Returns
    ----------
    pd.Series object that represents all the input genomes in the directory
    """

    genomes_ids = []
    genomes_pgfams = []

    with os.scandir(dir_path) as entries:
        for entry in entries:
            if entry.is_file():
                for extension, pgfam_col in PATRIC_FILE_EXTENSION_TO_PGFAM_COL.items():
                    if entry.name.endswith(extension):
                        genome_id = entry.name.split(extension)[0]
                        pgfams = read_genome_file(entry, pgfam_col)

                        genomes_ids.append(genome_id)
                        genomes_pgfams.append(pgfams)
                        break

    return pd.Series(genomes_pgfams, index=genomes_ids, dtype="string")


def read_genomes(path):
    """
    Reads all genomes information from an input directory with genome *.txt/*.tab files or a merged input *.fasta file

    Parameters
    ----------
    path - a path to an input directory with genome *.txt/*.tab files or a merged input *.fasta file

    Returns
    ----------
    pd.Series object that represents all the input genomes in the directory
    """

    if os.path.isdir(path):
        return read_files_in_dir(path)
    elif os.path.isfile(path):
        return read_merged_file(path)


def read_labels(path):
    """
    Reads csv file with labels from the given path

    Parameters
    ----------
    path -  path to *.csv file with labels

    Returns
    ----------
    labels - series object with the genomes labels
    """

    label_to_int = {HP: 1, NHP: 0, '1': 1, '0': 0}

    labels_df = pd.read_csv(path, dtype=str).set_index(GENOME_ID)
    labels = labels_df[LABEL].apply(lambda label: label_to_int.get(label.upper(), -1))

    return labels


def load_model(model_path):
    """
    Loads existing model from a model_path

    Parameters
    ----------
    model_path - path to the model file

    Returns
    ----------
    loaded model
    """

    with open(model_path, 'rb') as f:
        return pickle.load(f)


def load_model_str(data_str):
    """
    Loads existing model from data_str

    Parameters
    ----------
    data_str - pickled representation data of the model

    Returns
    ----------
    loaded model
    """

    return pickle.loads(data_str)



import argparse
import pickle

from os import path
from . import wspc
from datetime import datetime

PREDICT = 'predict'
FIT = 'fit'
TRAINED_MODEL_NAME = 'trained_model.pkl'
PREDICTIONS_FILE_NAME = 'predictions.csv'


def write_fitted_model(model, output_path):
    """
    Writes fitted model to pickle file

    Parameters
    ----------
    model - fitted model
    output_path - path to output directory
    """

    model_path = path.join(output_path, TRAINED_MODEL_NAME)

    with open(model_path, 'wb') as file:
        pickle.dump(model, file)


def write_predictions(predictions, output_path):
    """
    Writes model predictions to csv file

    Parameters
    ----------
    predictions - dataframe of the model predictions
    output_path - path to output directory
    """ 
    pred_path = path.join(output_path, PREDICTIONS_FILE_NAME)
    predictions.to_csv(pred_path)


def predict(args):
    """
    Predicts pathogenicity potentials 
    
    Parameters
    ----------
    args.i - input directory with genome *.txt files or a merged input *.fasta file
    args.model_path - path to a saved model in a *.pkl file. If not provided, saved pre-trained model will be used
    args.output - output directory, default current directory
    """

    X = wspc.read_genomes(args.i)

    model = wspc.load_model(args.model_path)

    predictions = wspc.predict(X, model)
    write_predictions(predictions, args.output)


def fit(args):
    """
    Fits new model 
    
    Parameters
    ----------
    args.i - path to input directory with genome *.txt files or a merged input *.fasta file
    args.labels_path - path to *.csv file with labels
    args.k - parameter for training - selecting k-best features using chi2
    args.t - parameter for training - clustering threshold
    args.output - output directory, default current directory
    """

    if not args.labels_path:
        raise ValueError('Please provide a path to labels using --labels_path')

    print(f"{datetime.now().strftime('%H:%M:%S')} Reading labels")
    y = wspc.read_labels(args.labels_path)
    print(f"{datetime.now().strftime('%H:%M:%S')} Read {y.shape[0]} labels")

    print(f"{datetime.now().strftime('%H:%M:%S')} Reading genomes")
    X = wspc.read_genomes(args.i, y)
    print(f"{datetime.now().strftime('%H:%M:%S')} Read {X.shape[0]} genomes")

    print(f"{datetime.now().strftime('%H:%M:%S')} Started training")
    model = wspc.fit(X, y, args.k, args.t)

    print(f"{datetime.now().strftime('%H:%M:%S')} Writing fitted model to {path.join(args.output, TRAINED_MODEL_NAME)}")
    write_fitted_model(model, args.output)


def main():

    parser = argparse.ArgumentParser()

    parser.add_argument('-m', '--mode', choices=[PREDICT, FIT], default=PREDICT)
    parser.add_argument('-i', required=True, help='input directory with genome *.txt/*.tab files or a merged '
                                                  'input *.fasta file')
    parser.add_argument('-o', '--output', help='output directory, default current directory', default='')
    parser.add_argument('-l', '--labels_path', help='path to *.csv file with labels')
    parser.add_argument('--model_path', help='path to a saved model in a *.pkl file. If not provided, saved pre-trained'
                                             ' model will be used')
    parser.add_argument('-k', type=int, help='parameter for training - selecting k-best features using chi2',
                        default=450)
    parser.add_argument('-t', type=float, help='parameter for training - clustering threshold', default=0.18)

    args = parser.parse_args()

    modes = {PREDICT: predict,
             FIT: fit}

    modes[args.mode](args)


if __name__ == "__main__":

    main()




import pandas as pd
import pkgutil
from . import feature_selection
from . import reader


PRED = 'Prediction'
PROB = 'Probability'
TRAINED_MODEL_PATH = 'model/WSPC_model.pkl'


def predict(X, model):
    """
    Predicts pathogenicity of input genomes represented by X

    Parameters
    ----------
    X - feature vectors of the input genomes
    model - trained model for the prediction

    Returns
    ----------
    results - pd.Dataframe object with the prediction results
    """

    y_pred, y_prob = model.predict(X), model.predict_proba(X)[:, 1]

    result = pd.DataFrame(data={PRED: y_pred, PROB: y_prob}, columns=[PRED, PROB], index=X.index)

    return result


def fit(X, y, k=450, threshold=0.18):
    """
    Fits classifier to an input training genomes

    Parameters
    ----------
    X - feature vectors of the input training genomes
    y - labels of the input training genomes

    Returns
    ----------
    pipeline - pd.Dataframe object with the prediction results
    """

    pipeline = feature_selection.get_fs_pipeline(k=k, threshold=threshold)
    pipeline.fit(X, y)

    return pipeline


def read_genomes(path, y=None):
    """
    Reads all genomes information from an input directory with genome *.txt files or a merged input *.fasta file

    Parameters
    ----------
    path - path to an input directory with genome *.txt files or a merged input *.fasta file
    y - if provided, reindex genomes according to y.index

    Returns
    ----------
    pd.Series object that represents all the input genomes in the directory
    """

    genomes = reader.read_genomes(path)
    if y is not None:
        genomes = genomes.reindex(y.index)

    return genomes


def read_labels(path, X=None):
    """
    Reads all genomes labels from *.csv file

    Parameters
    ----------
    path - path to *.csv file with labels
    X - if provided, reindex labels according to X.index

    Returns
    ----------
    labels - series object with the genomes labels
    """

    labels = reader.read_labels(path)

    if X is not None:
        labels = labels.reindex(X.index)
    return labels


def load_model(model_path=None):
    """
    Loads model from path to a saved model in a *.pkl file. If not provided, saved pre-trained model will be used

    Parameters
    ----------
    model_path - path to a saved model in a *.pkl file. If not provided, saved pre-trained model will be used

    Returns
    ----------
    loaded model
    """

    if not model_path:
        model_str = pkgutil.get_data(__name__, TRAINED_MODEL_PATH)
        return reader.load_model_str(model_str)

    return reader.load_model(model_path)




import numpy as np
import sklearn
import pandas as pd
import scipy.spatial.distance as ssd
from scipy.cluster import hierarchy
from scipy.stats import chi2_contingency
from sklearn.base import BaseEstimator
from sklearn.ensemble import RandomForestClassifier
from sklearn.feature_extraction.text import CountVectorizer
from sklearn.feature_selection import SelectKBest, SelectorMixin
from sklearn.pipeline import Pipeline


class SelectHierarchicalClustering(SelectorMixin, BaseEstimator):
    """
    A transformer that clusters the features in X according to dist_matrix, and selects a feature from each cluster with
    the highest chi2 score of X[feature] versus y
    """ 
    
    def __init__(self, dist_matrix=None, threshold=1):

        self.dist_matrix = dist_matrix
        self.threshold = threshold

    def _phi_coef(self, x, y):
        """
        Calculates phi coefficient between features

        Parameters
        ----------
        x - feature x column
        y - feature y column

        Returns
        ----------
        phi coefficient value
        """

        confusion_matrix = pd.crosstab(x, y)
        chi2 = chi2_contingency(confusion_matrix)[0]
        n = confusion_matrix.sum().sum()
        corr = np.sqrt(chi2 / n)

        return corr

    def _calc_dist_matrix(self, X):
        """
        Calculate distance matrix between each two features in X, each value is 1-phi_correlation
        """

        X_df = pd.DataFrame.sparse.from_spmatrix(X)
        X_corr_mat = X_df.corr(method=self._phi_coef)

        feature_corr_dist_matrix = 1 - X_corr_mat

        feature_corr_dist_matrix_condensed = ssd.squareform(feature_corr_dist_matrix)

        self.dist_matrix = feature_corr_dist_matrix_condensed

    def _corr_linkage(self, method='average'):

        linkage = hierarchy.linkage(self.dist_matrix, method=method)

        return linkage

    def _hierarchical_clustering(self, linkage):
        """ 
        Perform hierarchical clustering

        Parameters
        ----------
        linkage - linkage dendogram created by hierarchy.linkage(self.distance_matrix, method=method)

        Returns
        ----------
        a list of lists, each list represents a cluster and contains the indexes of features belonging
                 to the cluster
        """

        # array of len(X) - array[i] is the cluster number to which sample i belongs
        cluster_ids = hierarchy.fcluster(linkage, self.threshold, criterion='distance')

        cluster_id_to_feature_idx = {}
        for idx, cluster_id in enumerate(cluster_ids):
            cluster_id_to_feature_idx.setdefault(cluster_id, []).append(idx)

        return list(cluster_id_to_feature_idx.values())

    def fit(self, X, y):
        """
        Clusters the features (X columns) using self.dist_matrix and self.threshold, and selects a feature from each
        cluster with the highest chi2 score versus y.
        The attribute self.n_features_ represents the number of features selected (=number of clusters)
        The attribute self.selected_features_ is a list of indexes that correspond to the selected features
        """

        if not self.dist_matrix:
            self._calc_dist_matrix(X)

        linkage = self._corr_linkage()
        clusters = self._hierarchical_clustering(linkage)

        chi2_vals, __ = sklearn.feature_selection.chi2(X, y)
        chi2_vals = pd.Series(chi2_vals)

        # fitted attributes
        self.n_features_ = X.shape[1]
        self.selected_features_ = [chi2_vals[cluster].idxmax() for cluster in clusters]
        self.clusters_ = clusters

        print(f'threshold={self.threshold:.2f}, selected_features={len(self.selected_features_)}')

        return self

    def _get_support_mask(self):
        """
        Get the boolean mask indicating which features are selected

        Returns
        ----------
        mask - boolean array of shape [# input features]
            An element is True iff its corresponding feature is selected for
            retention.
        """

        # Checks if the estimator is fitted by verifying the presence of fitted attributes (ending with a trailing
        # underscore) and otherwise raises a NotFittedError with the given message.
        sklearn.utils.validation.check_is_fitted(self)

        mask = np.zeros((self.n_features_, ), dtype=bool)

        mask[self.selected_features_] = 1

        return mask


def get_fs_pipeline(k, threshold, random_state=0):
    """
    Creates feature selection pipeline

    Parameters
    ----------
    k - the k parameter for the SelectKBest features function
    threshold - clustering threshold for the Hierarchial clustering
    random_state - random state for the RandomForestClassifier. Deafult value: 0

    Returns
    ----------
    pipeline - feature selection pipeline
    """

    pipeline = Pipeline(steps=[('vectorize', CountVectorizer(lowercase=False, binary=True)),
                               ('k_best', SelectKBest(score_func=sklearn.feature_selection.chi2, k=k)),
                               ('cluster', SelectHierarchicalClustering(threshold=threshold)),
                               ('rf', RandomForestClassifier(random_state=random_state))])

    return pipeline



import setuptools

with open("README.md", "r", encoding="utf-8") as fh:
    long_description = fh.read()

setuptools.setup(
    name="wspc",
    version="0.0.6",
    author="Shaked Naor-Hoffmann & ‪Dina Svetlitsky",
    author_email="zivukelson@gmail.com",
    description="Protein-content-based bacterial pathogenicity classifier",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/shakedna1/wspc_rep",
    project_urls={
        "Bug Tracker": "https://github.com/shakedna1/wspc_rep/issues",
    },
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
    ],
    package_dir={"": "src"},
    packages=setuptools.find_packages(where="src"),
    python_requires=">=3.6",
    install_requires=[
        'pandas>=1.2.4',
        'numpy>=1.20.3',
        'scikit-learn==0.24.2',
        'scipy>=1.6.3'
    ],
    include_package_data=True,
    package_data={
        "wspc": ["src/wspc/model/WSPC_model.pkl"],
    },
    entry_points={
        'console_scripts': [
            'wspc = wspc:command_line.main',
        ],
    },
)
